from .constants import *

# Setting the display function
# This gets changed for Jupyter Notebook/IPython sessions, so that DataFrames
# are displayed in a fancier format
try:
    from IPython.display import display as disp
except:
    disp = print

class CostError(Exception):
    '''A custom exception for catching common input errors.

    In addition to error-type specific messages, this also captures and
    optionally displays the offending portions of the input DataFrames. 

    Parameters
    ----------
    message : str
        The error message to display. See `err_lines` dictionary for the
        custom messages.

    df : DataFrame (None)
        The bad portion of the input DataFrame.

    disp_df : Boolean (False)
        A flag that determines whether the bad DataFrame information is
        displayed along with the error message. Although this is useful for
        debugging, this parameter is set to False by default, because this
        could print sensitive information to the error logs.


    Notes
    -----
    See the `err_lines` dictionary defined below for the text of the various
    error messages. See the `CoreCost._sanity_check` method for the checks
    that are run.
    '''
    def __init__(self, message, df=None, disp_df=False):
        self.message = message
        self.df = df
        self.disp_df = disp_df
        super().__init__(message)

    def __str__(self, ):
        if self.disp_df:
            self.message += '\nSee the following entries:\n'
            self.message += self.df.to_string()
        return self.message


# A dictionary containing custom error messages for common input errors
err_lines = {
    'miss_mw': 'Missing MW error! \n'\
            'This most commonly happens for 1 of 2 reasons:\n'\
            '1. A reaction compound is not defined in the '\
            'materials table.\n'\
            '2. The compound names in reaction/materials '\
            'tables do not match *exactly* (even white space).',

    'dup_cpd': 'Duplicated material error!\n'\
            'Check that you do not have two or more compounds of the same\n'\
            'name in your material sheet(s).',

    'mis_cst': 'Missing material cost error!\n'\
           'This may be due to a missing material and/or its cost.\n',

    'mis_sol': 'Missing solvent info error!\n'\
           'One of your solvent entries is missing critical info.',
           
    }
    



import urllib.parse as parse
import warnings

import numpy as np
import pandas as pd

from .core import CoreCost
from .constants import *

# Filter warnings about data validation in Excel files
warnings.filterwarnings('ignore', category=UserWarning, 
                module='openpyxl')

# Setting the display function
# This gets changed for Jupyter Notebook/IPython sessions, so that DataFrames
# are displayed in a fancier format
try:
    from IPython.display import display as disp
    from IPython.display import Javascript
except:
    disp = print

# Set pandas to display lots of DataFrame rows so things don't get cut out
pd.options.display.max_rows = 1000
# For future reference -- Can't use this in conjunction with "fillna" becuase
# the DF doesn't display correctly 
# Set Pandas precision
#pd.set_option('precision', 2)

class ExcelCost(CoreCost):
    '''Costing class designed for reading data from Excel/csv spreadsheets.

    The Excel/csv files require specific columns to be present. See the
    `CoreCost` class doc string for more information.

    Parameters
    ----------
    materials_file : str
        The name/path of the materials list file. Can be xlsx or csv.

    rxn_file : str
        The name/path of the file defining the reactions. Can be xlsx or csv.
        Both `rxn_file` and `materials_file` can be the same file.

    final_prod : str
        The name of the final product for route. 

    materials_sheet : int, str (default = 0)
        The sheet in the Excel file that contains the materials information.
        The default (0) is the first sheet in the spreadsheet. You could use a
        different number or a name (str) if you want a different sheet from the
        spreadsheet.

    rxn_sheet : int, str (default = 0)
        See `materials_sheet` description, except this is for the sheet
        defining the reactions.

    alt_mat_file : str (default = None)
        The name/path of an optional, secondary materials sheet. This is
        useful if you have separate master and route-specific materials
        sheets, for example. Can be an xlsx or csv.

    alt_mat_sheet : int, str (default = 0)
        The sheet number/name for the secondary materials sheet. See
        `materials_sheet` description. 

    dis_err_df : boolean (default = False)
        If a `CostError` exception is raised, this controls whether the
        offending section of the DataFrame will be printed along with the
        error. This is useful for debugging, but is typically set to False, so
        that potentially sensitive information is not printed to the error
        logs.

    Notes
    -----
    See the `CoreCost` class doc string for additional information. 
    '''
    def __init__(self, materials_file, rxn_file, final_prod,
            materials_sheet=0, rxn_sheet=0, 
            alt_mat_file=None, alt_mat_sheet=0,
            disp_err_df=True, **kwargs):
        # Set up the reaction DataFrame
        self._rxn_file = rxn_file
        self._rxn_sheet = rxn_sheet
        rxns = self._rxn_read()

        # Create the Materials DataFrame from a main sheet and an optional
        # alternate sheet.
        self._materials_file = materials_file
        self._materials_sheet = materials_sheet
        self._alt_mat_file = alt_mat_file
        self._alt_mat_sheet = alt_mat_sheet
        materials = self._materials_build()                

        # Run the __init__ method from the CoreCost class
        super().__init__(materials, rxns, final_prod,
                disp_err_df=disp_err_df, **kwargs)

    def _rxn_read(self, ):
        '''Read an Excel sheet that defines the reactions.
        '''
        rxns = self._get_sheet_vals(self._rxn_file, self._rxn_sheet,
                            dtypes={RXN_STP:str,})
        return rxns

    def _get_sheet_vals(self, fname, fsheet, dtypes=None):
        '''A simple Excel/CSV reader function for both reaction and materials
        files.
        '''
        # Read the file, drop NaN-only and commented rows.
        if isinstance(fname, str) and fname[-3:].lower() == 'csv':
            df = pd.read_csv(fname, dtype=dtypes, comment='#')\
                            .dropna(how='all')
        else:
            df = pd.read_excel(fname, fsheet, dtype=dtypes, comment='#')\
                            .dropna(how='all')

        return df
        
    def _materials_build(self, ):
        '''Read and combine the main and an optional alternate materials
        sheets.
        '''
        materials = self._materials_read(self._materials_file,
                self._materials_sheet)

        # If an alternative materials key is given, combine that materials
        # sheet with the main one
        if self._alt_mat_file:
            alt_mats = self._materials_read(self._alt_mat_file,
                    self._alt_mat_sheet)
            # Concatenate the sheets. Reset the index so that it is
            # consecutively numbered
            materials = pd.concat([materials, alt_mats], sort=False)\
                    .reset_index(drop=True)

        # Set the final materials sheet
        return materials

    def _materials_read(self, mat_file, wsheet):
        '''Read an Excel file sheet that defines the materials used in
        costing.

        This is a separate method so it can be overwritten in other classes.
        '''
        mats = self._get_sheet_vals(mat_file, wsheet,)
        return mats

    def results(self, style='compact', decimals=2, fill='-'):
        '''Print the results of the costing calculation.

        Parameters
        ----------
        style : str, optional (Default = 'compact')
            This sets the style of the displayed costing DataFrame.
            `'compact'` prints a DataFrame that has been truncated slightly.
            `'full'` prints the entire DataFrame.

        decimals : int or None, optional (Default = 2)
            How many decimal places to show in the table. Set this to `None`
            if you want full precision.

        fill : str or None, optional (Default = '-')
            Fill NaN values with this string. This makes the table a little
            easier to read. Set this to `None` if you want to see the table
            with the typical NaN labels.
        '''
        # Get the prepared DataFrame for printing 
        fd = super().results(style)

        # Print the time the calculation was run
        utc_datetime = self._now.strftime('%Y-%m-%d %H:%M UTC')
        print(f'As of {utc_datetime} --')
        
        # Print a string about the final cost of the product
        cost_str = f'The final cost of {self.final_prod} is '\
                f'${self.cost:.2f}/kg.'
        print(cost_str)
            
        # Display the correct format of data based on the kwargs
        if isinstance(decimals, int):
            disp(fd.round(decimals).fillna(fill))
        else:
            disp(fd.fillna(fill))

    def excel(self, fname, ):
        '''Save the costing DataFrame as a dynamic Excel file.

        Parameters
        ----------
        fname : str, ExcelWriter
            The name you want to give to the Excel file, or an ExcelWriter
            object to send multiple costing sheets to the same file. See
            `pandas.DataFrame.to_excel` documentation for details.

        '''
        # Get the process DataFrame
        fd = super().excel()
       
        # Can't use a ":" in an Excel sheet name for time
        utc_datetime = self._now.strftime('%Y-%m-%d %Hh%Mm UTC')
        # Create the excel file. Can only save with the date and not the time
        fd.to_excel(fname, sheet_name=f'As of {utc_datetime}')
            

class ColabCost(ExcelCost):
    '''Costing class designed for the Colab Python environment.

    The materials and reaction information must be in Google Sheet format; the
    urls below are the URL values for those sheets. The user must have read
    access for the sheets to be able to use them. 

    Parameters
    ----------
    materials_url : str
        The URL to the Google sheet containing the materials list. 

    rxn_url : str
        The URL to the Google sheet containing the reaction information. 

    alt_mat_url : str, optional (default = None)
        The URL to the Google sheet containing an optional, secondary
        materials list.  This is useful if you have separate master and user
        materials sheets, for example.

    Notes
    -----
    This is subclass of `ExcelCost` and `CoreCost`, so see those class
    docstrings for more information.

    '''
    def __init__(self, materials_url, rxn_url, final_prod, materials_sheet=0,
            rxn_sheet=0, alt_mat_url=None, alt_mat_sheet=0, **kwargs):
        # Do some imports that are only possible in the Colab environment
        # This should prevent these from running in a non-Colab environment
        from google.auth import default
        from google.colab import auth
        from google.colab import files
        import gspread
        # These will have to be made global
        global default
        global auth
        global files
        global gspread

        # Authenticate the Colab environment 
        auth.authenticate_user()
        creds, _ = default()
        self._gc = gspread.authorize(creds)
        
        # Fix the final product and setup a mod variable
        super().__init__(materials_url, rxn_url, final_prod,
                materials_sheet, rxn_sheet, alt_mat_url, alt_mat_sheet, 
                **kwargs)
        
    def _materials_read(self, mat_url, wsheet):
        '''Read a Google sheet that defines the materials used in costing.

        Parameters
        ----------
        mat_key : str
            The unique key for a Google spreadsheet that defines the
            materials. 

        wsheet : str or int
            The specific sheet to extract from the Google spreadsheet. 
        '''
        mats = self._get_sheet_vals(mat_url, wsheet)

        # Warning for bad column namen
        self._col_warn(mats)

        # Convert numeric/date columns. Everything is read from a Google sheet
        # as strings
        num_cols = [MAT_MW, MAT_DEN, MAT_CST] 
        for nc in num_cols:
            mats[nc] = pd.to_numeric(mats[nc])

        return mats
        
    def _rxn_read(self, ):
        '''Read a Google Sheet containing reaction info and convert it to a
        DataFrame.
        '''
        rxns = self._get_sheet_vals(self._rxn_file,
                                     self._rxn_sheet)

        # Warning for bad column names
        self._col_warn(rxns)

        # Set some rxns columns to numeric values. Everything is read from a
        # Google sheet as strings
        num_cols = [RXN_EQ, RXN_VOL, RXN_RCY, RXN_OPX, RXN_MS]
        for nc in num_cols:
            if nc in rxns.columns:
                rxns[nc] = pd.to_numeric(rxns[nc])
        
        return rxns
        
    def _get_sheet_vals(self, key, sheet):
        '''General code for obtaining Google Sheet values and returning a 
        DataFrame.
        '''
        # Check if the `key` is a URL. If so, pull apart the url and grab the
        # part of the path that is the key to the spreadsheet.
        if key.startswith('http'):
            url_frag = parse.urlparse(key)
            key = url_frag.path.split('/')[3]

        # Using the Google sheet handle, pull down all values and make a 
        # DataFrame
        gsh = self._gc.open_by_key(key)
        # Differentiate between string and integer worksheets 
        if isinstance(sheet, str):
            ws = gsh.worksheet(sheet)
        else:
            ws = gsh.get_worksheet(sheet)
        vals = ws.get_all_values()
        val_df = pd.DataFrame(data=vals[1:], columns=vals[0])
        
        # Convert empty cells to NaN
        mask = (val_df == '')
        val_df[mask] = np.nan
        # Drop empty rows, these are particularly difficult to error check
        # sometimes. Assuming that both materials/rxns sheets should have
        # compound names for valid lines.
        cpd_mask = val_df[RXN_CPD].isna()
        val_df = val_df[~cpd_mask]

        # Drop lines with comment markers
        mask = ~val_df.iloc[:, 0].str.startswith('#')
        val_df = val_df[mask]
        
        return val_df

    def results(self, style='compact', decimals=2, fill='-'):
        '''Print the results of the costing calculation.

        See the `ExcelCost.results` method doc string for details on the
        function parameters.
        '''
        # This makes the max Colab output window very large, so that
        # DataFrames are not put into separate scroll windows, which is very
        # annoying for users. Unfortunately, I need to copy/paste the doc
        # string for results method, though...
        # See: https://github.com/googlecolab/colabtools/issues/541
        iframe_h = 'google.colab.output.setIframeHeight(0, true, {\n'\
                   '  maxHeight: 5000,\n'\
                   '})'
        disp(Javascript(iframe_h)) 
        # Call results method from ExcelCost
        super().results(style=style, decimals=decimals, fill=fill)

    def excel(self, fname):
        '''Download the costing DataFrame as an Excel file.

        Parameters
        ----------
        fname : str
            The name you want to give to the Excel file.

        '''
        super().excel(fname, )

        # Use the Colab function for downloading the file from the server
        files.download(fname)
        

class WebAppCost(ExcelCost):
    '''Costing class designed for the Streamlit Web App environment.

    This is largely the same as the `ExcelCost` class, so see that method for
    a more complete docstring. 
    '''
    def __init__(self, materials_file, rxn_file, final_prod,
            materials_sheet=0, rxn_sheet=0, 
            alt_mat_file=None, alt_mat_sheet=0,
            disp_err_df=False, **kwargs):

        super().__init__(materials_file, rxn_file, final_prod,
                materials_sheet, rxn_sheet, alt_mat_file, alt_mat_sheet,
                disp_err_df=disp_err_df, **kwargs)

    def results(self, style='compact', decimals=2, fill=np.nan):
        '''Returns the results of the costing calculation.

        See the `ExcelCost.results` method for additional documentation. This
        is largely the same as that method except for the optional `fill`
        keyword argument below.

        Parameters
        ----------
        fill : str, None, or Numpy.nan (Default = Numpy.nan)
            Fill NaN values with this string. This makes the table a little
            easier to read. Set this to `None` if you want to see the table
            with the typical NaN labels.
        '''
        # Use the CoreCost.results method to get the initial results DataFrame
        fd = super(ExcelCost, self).results(style)

        # Display the DataFrames for different permutations of kwargs
        if isinstance(decimals, int):
            fd = fd.round(decimals).fillna(fill)
        else:
            fd = fd.fillna(fill)

        return fd





'''Constants that are used across different modules.
'''
# Column name mapper: This is setting variable names for many of the important
# DataFrame columns. 
# Reactions Table Columns
RXN_STP = 'Step' # The step labels
RXN_CPD = 'Compound' # The compound names column, same for materials table
RXN_EQ = 'Equiv' # Molar equivalents of reagent
RXN_MS = 'Mass' # Mass of material used, typically kg, just be consistent
RXN_VOL = 'Volumes' # Volumes of solvent L/kg of limiting reagent (usually)
RXN_RCY = 'Recycle' # Fractional amount of solvent that is recycled
RXN_CST = 'Cost step' # Step where cost is calculated
RXN_OPX = 'OPEX' # Operational expenditures in $/kg, only for rxn product
NOTES = 'Notes' # The notes column name, same for materials table
# Materials Table Columns
MAT_MW = 'MW' # Molecular weight
MAT_DEN = 'Density' # Density, only used for solvents
MAT_CST = '$/kg' # The estimated material price/cost ($/kg)
# Calculated columns
RXN_KG = 'kg/kg rxn' # kg of materials used per kg reaction/step product
RXN_RMC = 'RM cost/kg rxn' # Cost of a material to make 1 kg rxn product
RXN_RMP = '% RM cost/kg rxn' # % material cost per 1 kg rxn product
PRD_KG = 'kg/kg prod' # kg of material used per kg of route product
PRD_RMC = 'RM cost/kg prod' # Cost of a material to make 1 kg route prod
PRD_RMP = '% RM cost/kg prod' # % material cost per 1 kg route product
# Excel temporary column names, these are renamed on Excel export
# These are the same as the calculated columns except for the SUFFIX
SUFFIX = ' dyn'
RNUM = 'rnum' # Row number column for indexing Excel cells
DYN_CST = MAT_CST + SUFFIX # Dynamic product cost
DYN_RKG = RXN_KG + SUFFIX # dynamic kg/kg rxn
DYN_RRMC = RXN_RMC + SUFFIX # dynamic RM cost/kg rxn
DYN_RRMP = RXN_RMP + SUFFIX # dynamic %RM cost/kg rxn
DYN_PKG = PRD_KG + SUFFIX # dynamic kg/kg prod
DYN_PRMC = PRD_RMC + SUFFIX # dynamic RM cost/kg prod
DYN_PRMP = PRD_RMP + SUFFIX # dynamic %RM cost/kg prod


# Excel Column dictionary
# This dictionary will be used for creating dynamic excel sheets, for which
# I'll need to know the DataFrame<->Excel column name mapping
ECOLS = {RXN_STP:'A', RXN_CPD:'B', MAT_MW:'C', MAT_DEN:'D', MAT_CST:'E',
        RXN_EQ:'F', RXN_VOL:'G', RXN_RCY:'H', RXN_CST:'I',
        RXN_OPX:'J', RXN_KG:'K', RXN_RMC:'L', RXN_RMP:'M', PRD_KG:'N',
        PRD_RMC:'O', PRD_RMP:'P', }

### References ###
# https://stackoverflow.com/questions/1383239/can-i-use-init-py-to-define-global-variables



from .constants import *

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

def value_scan(model, cpd, start, stop, npts, val_type='$/kg', step=None):
    '''Scan a range of values for a given material.
    
    Parameters
    ----------
    model : CoreCost or subclass
        This is the costing model instance that will be modified.
        
    cpd : str
        This the compound name for which the value will be modified.

    start : int, float
        The starting value for the scan.

    stop : int, float
        The ending value for the scan.

    npts : int
        The numbers of points to calculate the costs between `start` and
        `stop`

    val_type : str (Default = '$/kg')
        This is the column name for the parameter that you'll be changing.
        This must be for a non-calculated column, such as '$/kg', 'Equiv',
        'OPEX', etc.

    step : None, int (Default = None)
        The reaction step number for which this value will be changed. If
        this is `None` (default), then all the values for the given
        compound (`cpd`) will be set to the same value. This is mostly
        important for something like `val_type`='Equiv'. Clearly, you
        would only want to change the number of equivalents for a specific
        reaction. If this parameter is left as `None`, the equivalents for
        a given compound in all reactions will be set to the same value.

    Returns
    -------
    Pandas DataFrame
        This DataFrame has 'Values' and '$/kg' columns for the input
        values and output costs, repectively. 

    Notes
    -----
    Although this method recalculates the cost for every value, it does
    not modify the original `fulldata` or `cost` attributes. 
    '''
    # Create the values array
    vals = np.linspace(start, stop, npts)
   
    # I need a copy of the full data set in order to reset for each
    # iteration. Using this rather than rxn_data_setup because this makes it
    # possible to use a modified model with this function
    fd_copy = model.fulldata.copy()
    all_costs = []
    for val in vals:
        model.value_mod(cpd, val, val_type, step)
        model.calc_cost()
        all_costs.append(model.cost)
        # Reset the full data set 
        model.fulldata = fd_copy.copy()
        # Pop out the mod value, otherwise this list will get really long
        model._mod_vals.pop()

    # Reset the final cost
    model.cost = model.fulldata.loc[(model._fp_idx, model.final_prod), 
                              'RM cost/kg rxn']
    
    return pd.DataFrame({'Values':vals, '$/kg':all_costs})


def plot_scan(model, cpd, start, stop, npts, val_type='$/kg', step=None, 
            label=None):
    '''Plot a range of values.

    Parameters
    ----------
    See `value_scan` method description, except for the following.

    label : None, bool, str (Default = None)
        This will add a legend label. If you use the value `True`, the
        compound name will be added to the legend. Otherwise, you can pass a
        custom string if you want that to be in the legend instead. If None
        (default) or False is used, then no legend entry will be given. This
        function does add the legend to the plot; use pyplot.legend as a
        separate function call.
    '''
    # Calculate all the costs for the given values
    costs = value_scan(model, cpd, start, stop, npts, val_type=val_type,
                        step=step)
    
    # If label is boolean, make sure the label is set properly
    if label == True:
        label = cpd
    elif label == False:
        label = None

    # Plot the values
    plt.plot(costs['Values'], costs['$/kg'], 'o', label=label)


def sensitivity(model, col='Equiv', frac=0.1, decimals=2):
    '''Do a sensitivity analysis for the equivalents of reagents.

    Parameters
    ----------
    model : CoreCost or subclass
        This is the costing model instance that will be modified.
        
    col : str (Default = 'Equiv')
        Which column from the `fulldata` DataFrame should be used for the
        sensitivity analysis. 

    frac : float (Default = 0.1)
        Fractional percentage to increase/decrease the values by before
        recosting. The default is 0.1, which is the same as +/- 10%. 

    decimals : int or None (Default = 2)
        How many decimal places to display. If `None`, the full precision
        DataFrame will be displayed.
    '''
    # Make a new DF for sensitivity analysis. If material costs are being
    # screened, then remove the compounds being costed
    if col == MAT_CST:
        mask = model.fulldata[RXN_CST].isna()
        sens = model.fulldata.loc[mask, [col]]
    else:
        sens = model.fulldata[[col]].dropna()
    # Make values that are a certain percent above and below the current
    # numbers
    sens['Val low'] = sens[col]*(1 - frac)
    sens['Val high'] = sens[col]*(1 + frac)
    
    # Re-run the costing under the current conditions, which resets the
    # cost and fulldata variables. 
    model.calc_cost()
    # Make copies of these values so they don't change
    cost_save = model.cost
    fd_save = model.fulldata.copy()

    # Loop through the values and calculate the cost if these values
    # increase or decease by the percent given
    for step_cpd, vals in sens.iterrows():
        step, cpd = step_cpd
        # Low values
        model.value_mod(cpd, vals['Val low'], val_type=col, step=step)
        model.calc_cost()
        cost_low = model.cost
        cost_low_per = (cost_low*100./cost_save) - 100
        # Just to be safe - Drop the modified value
        model._mod_vals.pop()

        # High values
        model.value_mod(cpd, vals['Val high'], val_type=col, step=step)
        model.calc_cost()
        cost_high = model.cost
        cost_high_per = (cost_high*100./cost_save) - 100
        model._mod_vals.pop()

        # Set the values in the sensitivity DataFrame
        sens.loc[(step, cpd), 'Cost low'] = cost_low
        sens.loc[(step, cpd), 'Cost high'] = cost_high
        sens.loc[(step, cpd), '% low'] = cost_low_per
        sens.loc[(step, cpd), '% high'] = cost_high_per

        # Reset the original values.  This is necessary so the next step
        # in the loop starts from the original position.
        model.cost = cost_save
        model.fulldata = fd_save.copy()
    
    if decimals:
        return sens.round(decimals)
    else:
        return sens





import pytest
import pandas as pd
import numpy as np

# Using some of the testing routines from third-party libraries
from numpy.testing import *
from pandas.testing import *

from costcalc import CoreCost
from costcalc.constants import *
from costcalc.exceptions import CostError
from .data_build import *

### Tests - Working reactions ###

class Test_CoreFunctions(object):
    '''Tests for the public functions in the Core class.

    The test names here incorporate the public function name. These checks are
    parameterized with generic linear (lin) and convergent (con) routes, which
    is the minimal types of routes.  
    '''
    @pytest.mark.parametrize(
            "rxn, mat, fd",
            [(lin_clean, mat_clean, lin_clean_fd),
            (con_clean, mat_clean, con_clean_fd),
            (lin_split, mat_clean, lin_split_fd),
            (lin_opex, mat_clean, lin_opex_fd),
            (lin_mass, mat_clean, lin_mass_fd),
            ]
    )
    def test_rxn_data_setup(self, rxn, mat, fd):
        coster = CoreCost(mat, rxn, 'Product')
        assert_frame_equal(coster.fulldata, fd)

    @pytest.mark.parametrize(
            "rxn, mat, cost",
            [(lin_clean, mat_clean, 66.40818831355247,),
            (con_clean, mat_clean, 62.9051883439257,),
            (lin_split, mat_clean, 66.40818831355247,),
            (lin_opex, mat_clean, 112.41841531629551,),
            (lin_mass, mat_clean, 66.40818831355247,),
            ]
    )
    def test_calc_cost_clean(self, rxn, mat, cost):
        coster = CoreCost(mat, rxn, 'Product')
        coster.calc_cost()
        assert_allclose(coster.cost, cost)

    @pytest.mark.parametrize(
            "rxn, mat, res, style",
            [(lin_clean, mat_clean, lin_clean_res_c, 'compact'),
            (con_clean, mat_clean, con_clean_res_c, 'compact'),
            (lin_clean, mat_clean, lin_clean_res_f, 'full'),
            (con_clean, mat_clean, con_clean_res_f, 'full'),
            (lin_split, mat_clean, lin_split_res_f, 'full'),
            (lin_opex, mat_clean, lin_opex_res_f, 'full'),
            (lin_mass, mat_clean, lin_mass_res_f, 'full'),
            ]
    )
    def test_results(self, rxn, mat, res, style):
        coster = CoreCost(mat, rxn, 'Product')
        coster.calc_cost()
        assert_frame_equal(coster.results(style=style), res)


    @pytest.mark.parametrize(
            "rxn, mat, excel",
            [(lin_clean, mat_clean, lin_clean_excel,),
            (con_clean, mat_clean, con_clean_excel,),
            (lin_split, mat_clean, lin_split_excel,),
            (lin_opex, mat_clean, lin_opex_excel,),
            ]
    )
    def test_excel(self, rxn, mat, excel):
        coster = CoreCost(mat, rxn, 'Product')
        coster.calc_cost()
        assert_frame_equal(coster.excel(), excel)


class Test_CoreErrors(object):
    '''Test for known errors in the Core class.

    These errors are all captured in the code, so these are forcing the known
    errors to make sure they are correctly found/diagnosed.
    '''
    def test_miss_mw(self, ):
        '''Test for a missing material.
        '''
        mat_clean_miss = mat_clean.copy().drop(0)
        txt_match = 'Missing MW error!'
        with pytest.raises(CostError, match=txt_match) as err:
            coster = CoreCost(mat_clean_miss, lin_clean, 'Product')
        assert err.value.df.index == ('1', 'Starting Material')

    def test_missing_cost(self, ):
        '''Test for a missing cost or cost calculation.
        '''
        mat_clean_miss = mat_clean.copy()
        mask = mat_clean_miss[RXN_CPD] == 'Reagent C'
        mat_clean_miss.loc[mask, MAT_CST] = np.nan 
        txt_match = 'Missing material cost'
        with pytest.raises(CostError, match=txt_match) as err:
            coster = CoreCost(mat_clean_miss, lin_clean, 'Product')
        assert err.value.df.index == ('3', 'Reagent C')

    def test_duplicate_mat(self, ):
        '''Test for duplicated compound in the materials table.
        '''
        mat_clean_miss = mat_clean.copy()
        first_cpd = mat_clean_miss.iloc[:1,:]
        mat_clean_miss = pd.concat([mat_clean_miss, first_cpd])
        txt_match = 'Duplicated material error'
        with pytest.raises(CostError, match=txt_match) as err:
            coster = CoreCost(mat_clean_miss, lin_clean, 'Product')
        assert err.value.df.iloc[0, 0] == 'Starting Material'

    def test_missing_sol_density(self, ):
        '''Test for a missing density for solvent.
        '''
        mat_clean_dup = mat_clean.copy()
        mask = (mat_clean[RXN_CPD] == 'Water')
        mat_clean_dup.loc[mask, MAT_DEN] = np.nan
        txt_match = 'Missing solvent info error!'
        with pytest.raises(CostError, match=txt_match) as err:
            coster = CoreCost(mat_clean_dup, lin_clean, 'Product')
        assert err.value.df.index == ('2', 'Water')





import warnings

import numpy as np
import pandas as pd

from .constants import *
from .exceptions import *

# There are a bunch of warnings because the index isn't sorted; this filters
# those out. However, these performance warnings aren't really a problem
# because the DataFrames here are relatively small.
warnings.simplefilter(action='ignore', category=pd.errors.PerformanceWarning)

class CoreCost():
    def __init__(self, materials, rxns, final_prod, disp_err_df=False,
                filter_vals=True):
        '''Core class for calculating route costs and PMI.

        Parameters
        ----------
        materials : DataFrame
            A DataFrame containing  materials properties. At a minimum, this
            table must include the materials used in the route; however, it
            can contain an arbitrary number of additional materials as well,
            as long as the names are unique. This DataFrame must contain the
            following columns:

                * Compound : str, the compound names.
                * MW : float, the compound molecular weights.
                * Density : float, the density (g/mL) of the compound.
                Although this column must be present, density values are only
                necessary for reaction solvents.
                * $/kg : float, the cost/price of the compound in $/kg. Can be
                left blank for unknown costs, like intermediates and final
                products
                * Notes : str, optional notes about the compound.

            As noted above, the `Cost` column can be left blank for materials
            that will have their costs calculated. However, placeholder prices
            can be put here, if desired. These prices will be ignored if the
            compounds RM cost is being calculated in the route.

        rxns : DataFrame
            A DataFrame containing reaction information. This DataFrame must
            contain the following columns:

                * Step : str, a unique identifier for each step such as "1" or
                "1a"
                * Compound: str, the compound names. The first compound in any
                given Step should always be the reference compound for the
                reaction (i.e.  limiting reagent, solvent volume relative
                mass). The final compound for any given Step should always be
                the product.
                * Equiv : float, the equivalents used for each compound, for
                reaction products, this value should be the fractional yield
                (e.g. 75% yield is 0.75 equiv)
                * Mass : float, (Optional), the mass of material used. This is
                used to calculate equivalents. If used, the first compound
                for each reaction must have a mass provided.
                * Volumes : float, volume equivalents of solvent (L/kg)
                * Recycle : float, fractional percentage of material that can
                be recycled. E.g. 75% of a solvent can be recycled = 0.75
                * OPEX : float (Optional) a $/kg-step charge that is added to
                the final raw-material cost of a reaction product. Is only
                valid for reaction products.

        final_prod : str
            Name of the final product in the route.

        disp_err_df : Boolean (False)
            If a `CostError` exception is raised, this controls whether the
            offending section of the DataFrame will be printed along with the
            error. This is useful for debugging, but is typically set to
            False, so that potentially sensitive information is not printed to
            the error logs.

        filter_vals : Boolean (True)
            This keyword argument acts as a switch for filtering out certain
            values from the final DataFrame after a cost calculation is
            performed. By default (True), values will be filtered to correct
            several calculated columns for the route; however, switching this
            value may be convenient for evaluating all of the data per
            reaction. WARNING! Although the cost of the final product will
            still be correct if this argument is False, the "/kg rxn" and "/kg
            prod" columns as well as the PMI values will not be correct for
            the route as a whole.

        Attributes
        ----------
        final_prod : str
            The name of the overall final product of this route.
            
        rxns : DataFrame
            A DataFrame describing the reactions defined for the route. 
            
        materials : DataFrame
            A DataFrame describing all of the known materials. This will contain
            all of the materials from both the `materials_file` and the
            `alt_mat_file`.
            
        cost : float
            The final cost of the described route. This value is only present
            after the `calc_cost` method is called. 
            
        fulldata : DataFrame
            A combined DataFrame containing all of the reaction, material, and
            costing related values for the given route. 

        pmi : DataFrame
            A DataGrame containing the PMI for each reaction and the overall
            route. There will be an extra column with names used for sorting
            purposes; this column can be ignored.

        Notes
        -----
        Some automated value correction is done on the input DataFrames, such
        as removing blank lines and extra whitespace in compound names.
        However, other common errors will raise `CostError` exceptions, and
        these problems will require manual error correction from the user.
        '''
        # We need to store the input values/DataFrames
        self.rxns = rxns
        self.materials = materials
        self.final_prod = final_prod        
        self.filter_vals = filter_vals
        self._disp_err_df = disp_err_df

        # Correct typical input problems. 
        self._problem_correct()

        # Combine the reaction/materials sheets, add new columns
        self.rxn_data_setup()

        # Look for other common errors in the input. Throw an error if found
        # These errors must be manually fixed by the user, which is why they
        # are not covered in problem correction method
        self._sanity_check()

    def _problem_correct(self, ):
        '''Correct typical input problems in the input DataFrames.

        These are problems that tend to cause problems before merging using
        `rxn_data_setup` method, so they are automatically modified here. 
        '''
        rxn = self.rxns.copy()
        mat = self.materials.copy()

        # Warn about bad column names
        self._col_warn(mat)
        self._col_warn(rxn)

        # Drop lines that are still empty, these cause all sorts of
        # problems. Assume rxn/materials tables should have compound names
        # for valid entries
        rxn = rxn[ ~rxn[RXN_CPD].isna() ]
        mat = mat[ ~mat[RXN_CPD].isna() ]

        # Remove white space from before/after these columns
        # This is another tricky problem because trailing white space for
        # example is very hard to notice
        rxn_col = [RXN_STP, RXN_CPD]
        rxn[rxn_col] = rxn[rxn_col].apply(lambda x: x.str.strip())
        mat_col = [RXN_CPD, ]
        mat[mat_col] = mat[mat_col].apply(lambda x: x.str.strip())

        self.rxns = rxn.copy()
        self.materials = mat.copy()

    def _col_warn(self, df):
        '''Throw a warning if deprecated column names are used in the
        input DataFrames, and change to correct column name.  

        This is set as a temporary method because it is used here and the
        frontends.py module. If this is removed, be sure to correct both
        places.
        '''
        if "Cost" in df.columns:
            err = 'The "Cost" column name is no longer valid; '\
                    f'changing column name to "{MAT_CST}".'
            warnings.warn(err)
            df.rename(columns={"Cost": MAT_CST}, inplace=True) 

        if "Sol Recyc" in df.columns:
            err = 'The "Sol Recyc" column is no longer valid; '\
                    f'changing column name to "{RXN_RCY}".'
            warnings.warn(err)
            df.rename(columns={"Sol Recyc": RXN_RCY}, inplace=True)

    def rxn_data_setup(self, ):
        '''Setup the full data set for the upcoming cost calculations. 
        
        This method merges the materials and reaction sheets into a combined
        DataFrame called `fulldata`. It also serves as a "reset" of sorts
        after a costing calculation, because it uses the input materials and
        reactions DataFrames to rebuild the `fulldata` table. This is
        important if you are modifiying values, for example, and you want to
        put the system back into its starting point. 
        '''
        # Setup the reaction connectivity matrix. This is done by looking at
        # each reaction product, and assigning all usages of that product to
        # the correct step number. First, create an empty column for the steps
        self.rxns[RXN_CST] = np.nan
        steps = self.rxns.groupby(RXN_STP)
        for step, rxn in steps:
            # The product is the last compound in a given, save the final
            # index
            rxn_prod = rxn[RXN_CPD].iloc[-1]
            if rxn_prod == self.final_prod:
                self._fp_idx = step
            # Find all instances of the same product being used
            mask = self.rxns[RXN_CPD] == rxn_prod
            # Create the new costing column, add 
            self.rxns.loc[mask, RXN_CST] = step

        # Create an OPEX column, if it does not exist
        if RXN_OPX not in self.rxns.columns:
            self.rxns[RXN_OPX] = np.nan

        # Merge the materials and reaction DataFrames. A few columns are
        # dropped, which are not necessary for calculations. The merge happens
        # on the rxns DataFrame ('right'), which means that missing materials
        # will be fairly obvious (no MW, e.g.).
        mat_keeps = [RXN_CPD, MAT_MW, MAT_DEN, MAT_CST, NOTES]

        rxn_keeps = [RXN_STP, RXN_CPD, RXN_EQ, RXN_VOL, RXN_RCY, RXN_CST, 
                RXN_OPX, NOTES]
        # Check if an RXN_MS column is present. This is used to calculate
        # equivalents. If this is present, add this column to our "keep" list
        amt = False
        if RXN_MS in self.rxns.columns:
            amt = True 
            rxn_keeps = rxn_keeps[:2] + [RXN_MS] + rxn_keeps[2:]

        fulldata = pd.merge(self.materials[mat_keeps], self.rxns[rxn_keeps],
                            on=RXN_CPD, how='right')\
                            .rename({'Notes_x': 'Material Notes',
                                'Notes_y': 'Reaction Notes'}, axis=1)

        # If the RXN_MS column is present, then you will need calculate the
        # "Equiv" based on the mass given. Assumes that the first compound
        # is the limiting reagent.
        if amt:
            # Group everything by "Step" column
            grp = fulldata.groupby(RXN_STP)
            for idx, sb_grp in grp:
                # Grab the columns of interest, copy the ones to be modified
                masses = sb_grp[RXN_MS]
                equivs = sb_grp[RXN_EQ].copy()
                molwts = sb_grp[MAT_MW]
                volums = sb_grp[RXN_VOL].copy()
                densit = sb_grp[MAT_DEN]
                recycl = sb_grp[RXN_RCY]
                # If there are no amounts for a given Step, then skip
                if ~masses.any():
                    continue
                # If the first compound is not given as a mass,
                # raise an error
                if pd.isna(masses.iloc[0]):
                    err = f'Mass for limiting reagent in Step "{idx}" must '\
                            'be provided for masses to be used.'
                    raise ValueError(err)
                # Mask out only the values that have masses. This is
                # important in case a mixture of mass and equiv is used.
                mask = ~masses.isna()
                # Calculate the Equivalents
                equivs[mask] = masses[mask]/molwts[mask]
                # Assume the first entry is the limiting reagent. Normalize
                # the rest of the equivalents to that value
                equivs[mask] = equivs[mask]/equivs.iloc[0]
                # Set the values in the full DataFrame
                mask = fulldata[RXN_STP] == idx
                fulldata.loc[mask, RXN_EQ] = equivs
            # Remove the Amount column
            fulldata.drop(RXN_MS, axis=1, inplace=True)

        # Set MultiIndex
        fulldata.set_index([RXN_STP, RXN_CPD], inplace=True)
        
        # Save the full data set
        self.fulldata = fulldata

        # Reset the cost variable
        self.cost = None

        # Add a modified variable placeholder. This will store modified
        # values for the helper functions
        self._mod_vals = []

        # Add the empty columns
        self._column_clear()

    def _column_clear(self, excel=False):
        '''Clear out calculated values.

        This method will reset all the calculated values in the `fulldata`
        DataFrame. This is perhaps not strictly necessary, but it should help
        to avoid unwanted errors in recalculations due to old data still being
        present. This is not the same as a reset, though, because manually
        modified values with `value_mod` method will be re-modified. 
        '''
        # Set the costs to NaN for materials that will have costs calculated 
        cost_recalc_mask = ~self.fulldata[RXN_CST].isna()
        self.fulldata.loc[cost_recalc_mask, MAT_CST] = np.nan

        # Create or clear a bunch of columns that will be populated during 
        # cost calculation. 
        empty_cols = [RXN_KG, RXN_RMC, RXN_RMP, PRD_KG, PRD_RMC, PRD_RMP,]
        self.fulldata[ empty_cols ] = np.nan

        # For dynamic Excel
        if excel or self.fulldata.columns.str.contains(SUFFIX).any():
            excel_cols = [DYN_CST, DYN_RKG, DYN_RRMC, DYN_RRMP, DYN_PKG,
                    DYN_PRMC, DYN_PRMP,]
            self.fulldata[ excel_cols ] = ''

        # Change modified values that were set by helper functions
        for mod in self._mod_vals:
            self._set_val(*mod)

    def _sanity_check(self, ):
        '''Run some sanity checks on the DataFrames to catch common errors.

        Any problems found here will raise custom `CostError` exceptions. (See
        `costcalc.expections` module for the `CostError` expection class and
        error text.) Some of the errors are tricky, and the typical Python
        error output is unhelpful. These errors will require user
        intervention, unlike the issues corrected by the `_problem_correct`
        method. 
        ''' 
        # Check for a missing material -- Everything should have a MW
        # If not, this may suggest that there is a line that should not be
        # empty
        mw_mask = self.fulldata[MAT_MW].isna()
        if mw_mask.any():
            err_line = err_lines['miss_mw'] 
            df = self.fulldata[mw_mask].iloc[:,:3]
            raise CostError(err_line, df, self._disp_err_df)

        # Check for duplicated materials. This will probably be a big issue
        # with two materials sheets.
        dup_cpds = self.materials[RXN_CPD].duplicated()
        if dup_cpds.any():
            err_line = err_lines['dup_cpd'] 
            df = self.materials.loc[dup_cpds,].iloc[:, :3]
            raise CostError(err_line, df, self._disp_err_df)

        # Check for a missing cost, which is not being calculated
        # This is a tricky error because the costing will run just fine with 
        # NaNs. Check for this by looking for NaN in both Cost and Calc columns
        cost_mask = (self.fulldata[RXN_CST].isna() & \
                    self.fulldata[MAT_CST].isna())
        if cost_mask.any():
            err_line = err_lines['mis_cst']
            df = self.fulldata.loc[cost_mask, [MAT_CST, RXN_CST]]
            raise CostError(err_line, df, self._disp_err_df)
        
        # Check that all the solvent information is given
        sol_cols = [MAT_DEN, RXN_VOL]
        sol_mask = ~self.fulldata[RXN_VOL].isna() 
        sol_chk = self.fulldata.loc[sol_mask, MAT_DEN].isna()
        # If anything is missing, print a note
        if sol_chk.any():
            err_line = err_lines['mis_sol']
            sol_mask2 = sol_mask & sol_chk
            df = self.fulldata.loc[sol_mask2, sol_cols]
            raise CostError(err_line, df, self._disp_err_df)
       
    def _val_extract(self, df, idx, col):
        '''A convenience function for pulling a value from a DataFrame.

        In some circumstances, when you try to pull a single value, you get a
        Series instead. This just pulls a single value (first one) if a Series
        is returned.
        '''
        val = df.loc[idx, col]
        if isinstance(val, pd.Series):
            val = val.iloc[0]
        return val

    def calc_cost(self, excel=False):
        '''Calculate the cost of the route. 

        This is the main costing method. This will fill in the `fulldata`
        DataFrame with the appropriate values, and it will set a new class
        attribute `cost`, which is the total RM cost of the final product.

        Parameters
        ----------
        excel : boolean (False)
            If True, the calculations will output Excel-formated equation
            strings. By default (False), the output will be numerical values.
        '''
        # Save a time stamp so it can be displayed later
        self._now = pd.Timestamp.utcnow()
        # Prep the DataFrame
        self._column_clear(excel=excel)
        # Add an index of unique row labels to ensure that multiple entries of
        # the same compound name per row are handled correctly. This index
        # will need to be removed later
        nrows = self.fulldata.shape[0]
        self.fulldata['row'] = np.arange(nrows)
        self.fulldata.set_index('row', append=True, inplace=True)
        # For excel, there needs to be a string-based row numbering as well
        if excel:
            nrowdig = len(str(nrows))
            self.fulldata[RNUM] = [f'r{i:0{nrowdig:d}d}' for i in range(nrows)]
        # Run the costing and set the cost attribute
        self.cost = self._rxn_cost(self.final_prod, self._fp_idx, excel=excel)
        # Post process the DataFrame
        self._rxn_data_post(excel=excel)
        
    def _rxn_cost(self, prod, step, amp=1.0, eamp='', excel=False):
        '''The recursive cost calculating function. 
        
        This is the workhorse function of the whole process, but is not meant
        to be called on its own. Typically, you'll want to call `calc_cost` to
        get the costing information.
        
        Parameters
        ----------
        prod : str
            The name of the product for a particular reaction.

        step : str
            The step ID for which to calculate the cost. Because the step
            labels can be numbers (e.g. step 1) and/or letters (e.g. step 1a),
            this parameter is handled as a string.

        amp : float (default = 1.0)
            This number is an amplifier that increases some of the values,
            e.g. masses of materials, based on how much material is being used
            to make 1 kg of the overall final product.   

        eamp : str (default = '')
            This is the string amplifier value to use when creating dynamic
            Excel output.

        excel : boolean (False)
            Whether to output numerical values (False) or Excel-formatted
            equation strings (True).

        Notes
        -----
        The simplified algorithm here is as follows::

            Calculate kg of solids
            Calculate kg of solvents
            Normalize kg values
            For missing $/kg values:
                _rxn_cost(missing_compound, amp=kg_needed_for_miss_cpd)
            Calculate RMC and % for reaction
            Calculate RMC for product (multiply by amp)
            Return reaction product RMC

        The default behavior is to calculate the costs and other values as
        floating point numbers. However, there is also an Excel version here
        that returns the calculated values as equation strings, which will be
        executed when the table is opened in Excel. This uses an `ECOL`
        dictionary to map DF column names to Excel column letters, and an
        `RNUM` column that is the expected row numbers for the Excel output.
        '''
        # Select out the reaction of interest from the full data set. Saves
        # some typing. (Make this a copy to enusre it isn't given as a view.)
        data = self.fulldata.loc[step].copy()

        # Kg of nonsolvent materials used per equivalent
        data[RXN_KG] = data[RXN_EQ]*data[MAT_MW]
        # For Excel, normalize the data here. This will get overwritten for
        # solvents. 
        if excel:
            prod_row = self._val_extract(data, prod, RNUM)
            # Essentially "=Equivs*MW"
            data[DYN_RKG] = '=' + ECOLS[RXN_EQ] + data[RNUM] + '*'\
                    + ECOLS[MAT_MW] + data[RNUM]
            # Essentially above + "/(Prod_eq*Prod_MW)
            data[DYN_RKG] += '/(' + ECOLS[RXN_EQ] + prod_row + '*'\
                    + ECOLS[MAT_MW] + prod_row + ')'

        # Amount of solvent
        # First figure out which materials are solvents 
        mask = ~data[RXN_VOL].isna()
        sols = data[mask].copy()
        if sols.size != 0:
            # Find the material mass that the volumes are relative, which
            # should be for the first compound in the reaction
            amt_rel = data[RXN_KG].iloc[0]
            if excel:
                amt_rel_e = data[RNUM].iloc[0]
            # Calculate the mass of solvent. Take into account the solvent
            # recycyling 
            # kg sol = Volume*Density*(1-Recycle)*(kg SM)
            sols[RXN_KG] = sols[RXN_VOL]*sols[MAT_DEN]*amt_rel
            # And for Excel
            if excel:
                sols[DYN_RKG] = '=' + ECOLS[RXN_VOL] +\
                        sols[RNUM] + '*' + ECOLS[MAT_DEN] +\
                        sols[RNUM] + '*' + ECOLS[RXN_KG] +\
                        amt_rel_e
            data[mask] = sols

        # Set a recycle parameter, if given
        mask = ~data[RXN_RCY].isna()
        if mask.any():
            data.loc[mask, RXN_KG] *= (1 - data.loc[mask, RXN_RCY]) 
            if excel:
                dyn_rkg = data.loc[mask, DYN_RKG]
                # Wrap the original equation in parentheses
                dyn_rkg = '=(' + dyn_rkg.str.slice(1) + ')'
                # Multiply by the recycling parameter
                dyn_rkg += '*(1 - ' + ECOLS[RXN_RCY] +\
                            data.loc[mask, RNUM] + ')'
                data.loc[mask, DYN_RKG] = dyn_rkg

        # Normalize the kg of reaction
        prod_kg = self._val_extract(data, prod, RXN_KG)
        data[RXN_KG] /= prod_kg
        # Remove the 1 kg of product
        if self.filter_vals:
            data.loc[prod, RXN_KG] = np.nan
            if excel:
                data.loc[prod, DYN_RKG] = ''

        # Calculate unknown costs. 
        # Looks for any empty values in the "Cost" column. Don't use the
        # RXN_CST column directly, because some of the costs may have been
        # manually set using the `value_mod` method
        unknown_cost = data[MAT_CST].isna()
        # This is recursive. The final cost per kg of product will be
        # amplified by each subsequent step, which is where the new_amp
        # calculation comes into play
        for cpd_num, row in data.loc[unknown_cost].iterrows():
            cpd = cpd_num[0]
            # Don't do this for the product of the current reaction
            if cpd == prod: 
                continue
            # The amounts needed will be amplified by the appropriate kg ratio.
            # Set that ratio
            new_amp = self._val_extract(data, cpd, RXN_KG)
            # And for dynamic excel output
            if excel:
                new_amp_enum = self._val_extract(data, cpd, RNUM)
                kg_col = ECOLS[RXN_KG]
                # This will be '*(C#)'. 
                new_eamp = f'*({kg_col}{new_amp_enum})'

            # The new step is needed as well
            new_stp = self._val_extract(data, cpd, RXN_CST)
            # Run the cost calculation for the unknown compound
            if excel:
                cst = self._rxn_cost(cpd, new_stp, amp*new_amp, 
                                 eamp + new_eamp, excel=excel)
            else:
                cst = self._rxn_cost(cpd, new_stp, amp*new_amp)
            # Set the calculated cost
            data.loc[cpd, MAT_CST] = cst
            # And for Excel -- This cost will need to get swapped out later.
            # Also need to check if an OPEX is necessary
            if excel:
                eopex = self._val_extract(self.fulldata, (new_stp, cpd), 
                                        RXN_OPX)
                ernum = self._val_extract(self.fulldata, (new_stp, cpd), 
                                        RNUM)

                if np.isnan(eopex):
                    data.loc[cpd, DYN_CST] = '=' + ECOLS[MAT_CST] + ernum
                            
                else:
                    data.loc[cpd, DYN_CST] = '=' + ECOLS[MAT_CST] + ernum +\
                            '+' + ECOLS[RXN_OPX] + ernum

        # Calculate the cost for each material in the reaction
        data[RXN_RMC] = data[RXN_KG]*data[MAT_CST]
        # The product cost will be the sum of all the reactant/solvent costs
        prd_cost = data[RXN_RMC].sum()
        data.loc[prod, RXN_RMC] = prd_cost
        data.loc[prod, MAT_CST] = prd_cost
        # And for Excel
        if excel:
            data[DYN_RRMC] = '=' + ECOLS[RXN_KG] + data[RNUM] + '*' + \
                            ECOLS[MAT_CST] + data[RNUM]
            # And for Excel, first we need the rows that are not the product
            mask = data.index[:-1]
            # Then we need to make a comma-separated list of these rows
            cells = [f'{ECOLS[RXN_RMC]}{r}' for r in data.loc[mask, RNUM]]
            rs = ','.join(cells)
            # Combine them together into a sum
            data.loc[prod, DYN_RRMC] = '=SUM(' + rs + ')'
            # Set the cost to the calculated value
            data.loc[prod, DYN_CST] = '=' + ECOLS[RXN_RMC] + prod_row

        # Calculate % costs for individual rxn
        # = (RM cost/kg rxn)/(RM cost/kg rxn for the rxn product)
        prod_rmc = self._val_extract(data, prod, RXN_RMC)
        per_rm_cost = data[RXN_RMC]*100/prod_rmc
        data[RXN_RMP] = per_rm_cost.values
        # And for Excel
        if excel:
            data[DYN_RRMP] = '=' + ECOLS[RXN_RMC] + data[RNUM] +\
                    '*100/' + ECOLS[RXN_RMC] + prod_row
        # Remove the % cost for the rxn product
        if self.filter_vals:
            data.loc[prod, RXN_RMP] = np.nan
            # And for Excel
            if excel:
                data.loc[prod, DYN_RRMP] = ''

        # These are the costs for ultimate product
        # For one reaction amp=1, so the individual rxn cost = ultimate rxn 
        # cost. However, for feeder reactions this will get amplified by 
        # each step   
        data[PRD_RMC] = data[RXN_RMC]*amp
        # And for Excel
        if excel:
            data[DYN_PRMC] = '=' + ECOLS[RXN_RMC] + data[RNUM] + eamp
        
        # This sets the number of kg of each material per kilogram of product
        # This is done by multiplying the per reaction value by the amplifier
        # This isn't necessary for costing, but we can use it for PMI
        data[PRD_KG] = data[RXN_KG]*amp
        # And for Excel
        if excel:
            data[DYN_PKG] = '=' + ECOLS[RXN_KG] + data[RNUM] + eamp
        
        # Set the values in the big DataFrame with this slice.  
        self.fulldata.loc[step, data.columns] = data.values
        
        # Return the calculated product cost, which is required for the 
        # recursive nature of the algorithm. In addition, an optional OPEX
        # may be added to take into account production costs of the cpd
        opex = self._val_extract(data, prod, RXN_OPX)
        if np.isnan(opex):
            return self._val_extract(data, prod, RXN_RMC)
        else:
            return self._val_extract(data, prod, RXN_RMC) + opex
    
    def _rxn_data_post(self, excel=False):
        '''Calculate some values after the final cost of the reaction is
        determined. 

        This includes the final "% RM cost/kg prod", setting the final cost
        with an optional OPEX, and all PMI calculations. Some values are
        filtered out as well to make the column sums sensible.
        '''
        prod = self.final_prod
        step = self._fp_idx

        # Remove the "row" index
        self.fulldata.reset_index(level=2, drop=True, inplace=True)
        
        # If an OPEX for the final reaction is given, add that to the cost
        # of the final product
        opex = self._val_extract(self.fulldata, (step, prod), RXN_OPX)

        if not np.isnan(opex):
            self.fulldata.loc[(step, prod), MAT_CST] = self.cost
            # And for Excel
            if excel:
                prod_row = self._val_extract(self.fulldata, (step, prod), 
                        RNUM)
                self.fulldata.loc[(step, prod), DYN_CST] += '+' + \
                        ECOLS[RXN_OPX] + prod_row
                
        # Calculate % overall costs relative to the prod
        self.fulldata[PRD_RMP] = self.fulldata[PRD_RMC]*100/self.cost

        # And for Excel
        if excel:
            prod_row = self._val_extract(self.fulldata, (step, prod), RNUM)
            self.fulldata[DYN_PRMP] = '=' +\
                    ECOLS[PRD_RMC] + self.fulldata[RNUM] +\
                    '*100/' + ECOLS[RXN_RMC] + prod_row
                    
        
        # Filter out certain values to simplify full data set
        if self.filter_vals:
            # Remove the cost and %s for cost-calculated materials
            # This is necessary so that this column adds up to 100% (w/o OPEX)
            mask = ~self.fulldata[RXN_CST].isna()
            self.fulldata.loc[mask, PRD_RMP] = np.nan
            if excel:
                self.fulldata.loc[mask, DYN_PRMP] = ''

            # This filters some of the costs which are simply the sum of raw
            # materials from earlier rxns. The sum of this column will now be
            # equal to the cost of the final product.
            self.fulldata.loc[mask, PRD_RMC] = np.nan
            if excel:
                self.fulldata.loc[mask, DYN_PRMC] = ''

            # This filters out the kg/kg prod values that were calculated, so that
            # the sum of this column is the PMI
            self.fulldata.loc[mask, PRD_KG] = np.nan
            if excel:
                self.fulldata.loc[mask, DYN_PKG] = ''

        # PMI Calculations
        # Adding a prefix for display purposes
        # There will be a funny column in this DF with these values...
        self._pre = '*'
        
        # First of all, calculate the PMI for each reaction individually
        gb = self.fulldata.groupby(RXN_STP)
        if excel:
            rxn_pmi = gb.agg({RXN_KG:'sum', RNUM:self._excel_pmi})\
                            .rename({RNUM: DYN_RKG}, axis=1)\
                            .reset_index()
        else:
            rxn_pmi = gb.agg({RXN_KG:'sum'}).reset_index()

        rxn_pmi[RXN_CPD] = self._pre + 'Step ' + rxn_pmi[RXN_STP] + ' PMI'
        
        # The full route PMI is not the sum of the above, but is the sum of
        # the PRD_KG column. We need to make this into a DataFrame to
        # merge with the per reaction values above.
        df_vals = {PRD_KG: [self.fulldata[PRD_KG].sum()], 
                   RXN_STP: [self._fp_idx],
                   RXN_CPD: [self._pre*2 + 'Full Route PMI'],
                   }
        if excel:
            mask = ~self.fulldata[PRD_KG].isna()
            all_cells = [f'{ECOLS[PRD_KG]}{i}' for i in\
                         self.fulldata.loc[mask, RNUM]]
            cell_sum = '=SUM(' + ','.join(all_cells) + ')'
            df_vals[DYN_PKG] = [cell_sum]
        full_pmi = pd.DataFrame(df_vals)

        # Merge the per-reaction and full PMI
        self.pmi = pd.concat([rxn_pmi, full_pmi], sort=False)\
                .set_index(RXN_STP)

    def _excel_pmi(self, col):
        # A function for creating the dynamic Excel cells for PMI
        # Pull the Step value from the index
        step = col.index[0][0]
        # Mask out the reaction product, use mask values only to avoid 
        # indexng errors in the `cells` loop
        mask = (self.fulldata.loc[step, RXN_CST] != step).values
        cells = [f'{ECOLS[RXN_KG]}{i}' for i in col[mask]]
        return "=SUM(" + ','.join(cells) + ")"

    def results(self, style='compact'):
        '''Returns a formatted/simplified full output DataFrame.

        The `fulldata` attribute contains a number of potentially unnecessary
        columns that need to be filtered out for presentation purposes. This
        also combines the PMI data as well, so everything is in one table.

        Parameters
        ----------
        style : str ('compact')
            By default, this returns a minimal representation of the system
            ('compact'); however, a more complete set of columns can be
            returned if the value `'full'` is passed as the parameter here.
        '''
        if not self.cost:
            err = 'No cost calculation information. Use the `calc_cost`'\
                    ' method first.'
            raise RuntimeError(err)

        # For compact display, these are the most important columns
        comp_col = [MAT_CST, RXN_EQ,] 
        if self.fulldata[RXN_VOL].any():
            comp_col.extend([RXN_VOL, RXN_RCY,])
        if self.fulldata[RXN_OPX].any():
            comp_col.append(RXN_OPX) 
        comp_col.extend([RXN_KG, RXN_RMC, RXN_RMP, PRD_KG, PRD_RMC, PRD_RMP])

        # For full display, don't show the Excel columns
        ecol_mask = ~self.fulldata.columns.str.contains(SUFFIX+'|'+RNUM)
        no_ecol = self.fulldata.columns[ecol_mask]
        
        # Combine the fulldata and pmi DataFrames
        fd = self._df_combine()

        # Display the DataFrames for different permutations of kwargs
        # The fillna removes NaN from the displayed tables in Notebook, but
        # may goof up printing
        if style == 'full':
            fd = fd[no_ecol]
        elif style == 'compact':
            fd = fd[comp_col]

        return fd

    def _df_combine(self, ):
        '''Combine the fulldata and pmi DataFrames for saving/exporting.
        '''
        # Copy the original DFs, and remove indexing
        fd = self.fulldata.reset_index()
        pmi = self.pmi.reset_index()
        
        # Concats the fulldata and pmi based on step. The groupby function
        # will sort by "Step" as it processes things. This will need to be
        # modified if this behavior is undesired, but as is, it is not trivial
        # to make this fix. The apply function doubles up the "Step" column.
        # So one needs to be dropped.
        gb = fd.groupby(RXN_STP)
        concated = gb.apply(self.__fd_pmi_concat, pmi)
        concated = concated.drop(RXN_STP, axis=1)\
                            .reset_index()\
                            .drop('level_1', axis=1)
        
        # Reset the index and return the DF. No sorting is necessary here.
        return concated.set_index([RXN_STP, RXN_CPD])                

    def __fd_pmi_concat(self, df, pmi):
        '''Concats the fulldata and pmi DataFrames.

        This is only used in a DataFrame.apply call in the _df_combine
        method.'''
        # The step number
        step = df[RXN_STP].iloc[0]
        # Mask out the correct PMI values
        pmi_mask = pmi[RXN_STP] == step
        pmi_small = pmi[pmi_mask]
        # Combine the route step info with the PMI. In this way, the pmi will
        # be at the end.
        return pd.concat([df, pmi_small], ignore_index=True)

    def excel(self, ):
        '''Return an Excel-formatted DataFrame.

        This table will have equation strings rather than numerical values,
        and it can be saved to disk using the `DataFrame.to_excel` method, if
        desired. Because the table contains equation strings, the Excel file
        will be fully dynamic, in that value changes will cause a
        recalculation of the entire sheet.
        '''
        # Run the cost calculation again, but using the excel keyword
        self.calc_cost(excel=True)

        # Combine the fulldata and pmi DataFrames
        fd = self._df_combine()
        # Move the Notes columns to the end of the combined DataFrame
        mat_note = fd.pop('Material Notes')
        fd.insert(fd.shape[1], 'Material Notes', mat_note)
        rxn_note = fd.pop('Reaction Notes')
        fd.insert(fd.shape[1], 'Reaction Notes', rxn_note)

        # Convert the unique row ID with an Excel sheet row number
        nrows = fd.shape[0]
        for r, n in zip(fd[RNUM], np.arange(2, nrows+2)):
            if isinstance(r, float):
                continue
            for col in fd:
                if SUFFIX in col:
                    fd[col] = fd[col].str.replace(r, str(n))
                    
        # Set dynamic cost for the "Cost" of calculated products
        mask = fd[DYN_CST] != ''
        fd.loc[mask, MAT_CST] = fd.loc[mask, DYN_CST]
        
        # Drop the non-dynamic columns
        d_cols = [DYN_CST, RXN_KG, RXN_RMC, RXN_RMP, PRD_KG, PRD_RMC, PRD_RMP,
                RNUM]
        fd = fd.drop(d_cols, axis=1)
        
        # Rename columns to remove "dyn" SUFFIX
        col_str = fd.columns.str.replace(SUFFIX, '')
        new_cols = {c:cn for (c, cn) in zip(fd.columns, col_str)}
        fd = fd.rename(new_cols, axis=1)

        # Reset all "empty" cells to NaN. This is important for unit testing
        # purposes.
        fd = fd.where(fd != '', np.nan) 

        # Rerun the cost calculation without the excel stuff to get rid of all
        # the other columns
        self.calc_cost(excel=False)

        return fd

    def value_mod(self, cpd, val, val_type='$/kg', step=None):
        '''Manually set a value for a given material.

        This is useful for creating alternative models without having to
        import completely new reactions and materials tables. The original
        input tables are not modified, so the model can be reset with the
        `rxn_data_setup` method. 

        Parameters
        ----------
        cpd : str
            This the compound name for which the value will be modified.

        val : int, float
            This is the modified value for the parameter.

        val_type : str, optional (Default = '$/kg')
            This is the column name for the parameter that you'll be changing.
            This must be for a non-calculated column, such as '$/kg', 'Equiv',
            'OPEX', etc.

        step : None, int, str (Default = None)
            The reaction step number for which this value will be changed. If
            this is `None` (default), then all the values for the given
            compound (`cpd`) will be set to the same value. This is mostly
            important for something like `val_type`='Equiv'. Clearly, you
            would only want to change the number of equivalents for a specific
            reaction. If this parameter is left as `None`, the equivalents for
            a given compound in all reactions will be set to the same value.
        
        Note
        ----
        This method will *NOT* recalculate the cost; this must be done as a
        separate step. This behavior is by design so that several `value_mod`
        method calls can be made, if necessary, before the costs are
        recaclulated.

        In addition, if several entries of the same compound names are found
        in a particular reaction Step, they will all be modified equally.
        There is no way (currently) to separate these compounds. In principle,
        a user could make a new subclass of the CoreCost class or its
        subclasses to add an additional indexing level.
        '''
        # Store the values
        self._mod_vals.append( (cpd, val, val_type, step) )
        # This will clear out the old calculation data and set the modified
        # value. Keeps folks from getting confused b/c previously calculated
        # values appear unchanged.
        self._column_clear()

    def _set_val(self, cpd, val, val_type, step):
        '''Set a modified value in the `fulldata` DataFrame.

        The modified values are set using the `value_mod` method, which adds
        the values to a `_mod_vals` list. This function cycles through that
        list, changing the associated values.
        '''
        # The first one sets all values w/ the compound name. The second one
        # sets only a value for a specific reaction.
        if not step:
            cells = (slice(None), cpd)
        else: 
            # The step needs to be a string, not in int as might be expected
            cells = (str(step), cpd)
        
        # Does this position actually exist???
        try:
            self.fulldata.loc[cells, val_type]
        except KeyError:
            # If not, remove the values from the list and throw an error
            self._mod_vals.pop()
            err = f'Compound "{cpd}"'
            if step:
                err = f'Step "{step}", {err}'
            err += " does not exist. Check "
            err += "your reactions to find the correct Step/Compound."
            raise KeyError(err) 
            
        self.fulldata.loc[cells, val_type] = val
        # The "Cost calc" flag must be set to np.nan when setting a cost. 
        # This is necessary for % RM cost calcs, e.g.
        if val_type == MAT_CST:
            self.fulldata.loc[cells, RXN_CST] = np.nan


